package dal

import (
	"context"
	"crypto/rand"
	"database/sql"
	"encoding/binary"
	"fmt"
	"goblog/util"
	"log"
)

var ErrCreateUserExists = fmt.Errorf("user exists")

func CheckUserIdentity(username string, password string) (ok bool, err error, uid uint32) {
	type queryResult struct {
		Count int    `db:"count"`
		Id    uint32 `db:"id"`
	}
	var result queryResult
	err = db.Get(
		&result,
		"SELECT COUNT(*) as count, IFNULL(id, 0) as id FROM users WHERE username=$1 AND hashed_password=$2",
		username,
		util.HashPassword(password),
	)
	if err != nil {
		return
	}
	return result.Count == 1, nil, result.Id
}

func CreateUser(username string, password string) (uid uint32, err error) {
	tx, err := db.BeginTx(context.TODO(), &sql.TxOptions{Isolation: sql.LevelSerializable})
	if err != nil {
		err = fmt.Errorf("failed to start transaction: %w", err)
		return
	}

	rollback := true
	defer func() {
		if rollback {
			e := tx.Rollback()
			if e != nil {
				log.Printf("CreateUser: Failed to rollback: %v\n", e)
			}
		}
	}()

	query := tx.QueryRow("SELECT COUNT(*) FROM users WHERE username=$1", username)
	if err != nil {
		err = fmt.Errorf("failed to check username availability: %w", err)
		return
	}
	var count int
	err = query.Scan(&count)
	if err != nil {
		err = fmt.Errorf("failed to check username availability (scan): %w", err)
		return
	}
	if count > 0 {
		err = ErrCreateUserExists
		return
	}

	for {
		b := make([]byte, 4)
		_, err = rand.Read(b)
		if err != nil {
			err = fmt.Errorf("failed to read urandom: %w", err)
			return
		}
		uid = binary.LittleEndian.Uint32(b)

		query := tx.QueryRow("SELECT COUNT(*) FROM users WHERE id=$1", uid)
		if err != nil {
			err = fmt.Errorf("failed to check uid availability: %w", err)
			return
		}
		var count int
		err = query.Scan(&count)
		if err != nil {
			err = fmt.Errorf("failed to check uid availability (scan): %w", err)
			return
		}

		if uid > 1000 && count <= 0 {
			break
		}
	}

	result, err := tx.Exec(`INSERT INTO users (id, username, description, hashed_password, is_profile_public) 
		VALUES ($1, $2, $3, $4, $5)`, uid, username, "(empty)", util.HashPassword(password), 0)
	if err != nil {
		return
	}
	rows, err := result.RowsAffected()
	if err != nil {
		return
	}
	if rows <= 0 {
		err = ErrCreateUserExists
	}

	err = tx.Commit()
	if err != nil {
		err = fmt.Errorf("failed to commit: %w", err)
		return
	}
	rollback = false

	log.Println("Register: ", username, password, uid)
	return
}
